# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko, 2006 Allen Bierbaum, Matthias Baas
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines class, :class:`decl_printer_t` that prints declarations tree in a
user friendly format
"""

import os
import sys
from . import calldef_members
from . import algorithm
from . import decl_visitor
from . import variable_t
from . import calldef_t
from . import type_traits_classes


def _stdout_writer(x):
    sys.stdout.write(x)


class decl_printer_t(decl_visitor.decl_visitor_t):

    """helper class for printing declarations tree"""

    JUSTIFY = 20
    INDENT_SIZE = 4

    def __init__(
            self,
            level=0,
            print_details=True,
            recursive=True,
            writer=_stdout_writer,
            verbose=True):
        decl_visitor.decl_visitor_t.__init__(self)
        self.__inst = None
        self.__level = level
        self.__print_details = print_details
        self.__recursive = recursive
        self.__verbose = verbose
        self.__writer = writer

    def clone(self, increment_level=True):
        level = self.__level
        if increment_level:
            level += 1
        return decl_printer_t(
            level,
            self.print_details,
            recursive=self.recursive,
            verbose=self.__verbose,
            writer=self.writer)

    @property
    def recursive(self):
        return self.__recursive

    @recursive.setter
    def recursive(self, recursive):
        self.__recursive = recursive

    @property
    def verbose(self):
        return self.__verbose

    @verbose.setter
    def verbose(self, verbose):
        self.__verbose = verbose

    @property
    def level(self):
        return self.__level

    @level.setter
    def level(self, level):
        self.__level = level

    @property
    def print_details(self):
        return self.__print_details

    @print_details.setter
    def print_details(self, details):
        self.__print_details = details

    @property
    def writer(self):
        return self.__writer

    @writer.setter
    def writer(self, writer):
        self.__writer = writer

    @property
    def instance(self):
        return self.__inst

    @instance.setter
    def instance(self, inst):
        self.__inst = inst

    @staticmethod
    def is_builtin_decl(decl):
        if not decl.name.startswith('__builtin_'):
            return False

    @staticmethod
    def __nice_decl_name(inst):
        return inst.__class__.__name__

    def print_decl_header(self):
        header = self.__nice_decl_name(
            self.__inst) + ": '%s'" % self.__inst.name
        self.writer(
            ' ' *
            self.level *
            self.INDENT_SIZE +
            header.ljust(self.JUSTIFY) +
            os.linesep)
        if self.__print_details:
            curr_level = self.level + 1
            if self.__inst.location:
                location = 'location: [%s]:%s' % (
                    self.__inst.location.file_name, self.__inst.location.line)
                self.writer(
                    ' ' *
                    curr_level *
                    self.INDENT_SIZE +
                    location)
            if self.verbose:
                artificial = 'artificial: ' + \
                    "'%s'" % str(self.__inst.is_artificial)
                self.writer(
                    ' ' *
                    curr_level *
                    self.INDENT_SIZE +
                    artificial.ljust(self.JUSTIFY))
                if self.__inst.attributes:
                    attributes = 'attributes: %s' % self.__inst.attributes
                    self.writer(
                        ' ' *
                        curr_level *
                        self.INDENT_SIZE +
                        attributes)

                # Mangled name is only available for functions and variables
                # when using castxml.
                print_mangled = False
                if self.__inst.mangled and \
                    (isinstance(self.__inst, variable_t) or
                     isinstance(self.__inst, calldef_t)):
                    print_mangled = True

                if print_mangled:
                    mangled = 'mangled: %s' % self.__inst.mangled
                    self.writer(
                        ' ' *
                        curr_level *
                        self.INDENT_SIZE +
                        mangled)

    def print_calldef_info(self, decl=None):
        if None is decl:
            decl = self.__inst

        retval = None
        if decl.return_type:
            retval = decl.return_type.decl_string
        args = []
        for arg in decl.arguments:
            args.append(arg.decl_type.decl_string + ' ' + arg.name)
        indent = ' ' * (self.level + 1) * self.INDENT_SIZE
        self.writer(indent + "is extern: " + str(decl.has_extern))
        self.writer(indent + "return type: " + str(retval))
        self.writer(indent + "arguments type: " + ', '.join(args))
        self.writer(
            indent +
            "calling convention: __%s__" %
            decl.calling_convention +
            os.linesep)
        if isinstance(decl, calldef_members.member_calldef_t):
            self.writer(indent +
                        "virtual: " +
                        str(decl.virtuality) +
                        os.linesep)
            self.writer(indent +
                        "is const: " +
                        str(decl.has_const) +
                        os.linesep)
            self.writer(indent +
                        "is static: " +
                        str(decl.has_static) +
                        os.linesep)

    def visit_member_function(self):
        self.print_decl_header()
        self.print_calldef_info()

    def visit_constructor(self):
        self.print_decl_header()
        self.print_calldef_info()

        indent = ' ' * (self.level + 1) * self.INDENT_SIZE
        self.writer(indent +
                    "explicit: " +
                    str(self.__inst.explicit) +
                    os.linesep)

        if self.__print_details:
            self.writer(indent +
                        'copy constructor: ' +
                        str(type_traits_classes.is_copy_constructor(
                            self.__inst)) +
                        os.linesep)

    def visit_destructor(self):
        self.print_decl_header()

    def visit_member_operator(self):
        self.print_decl_header()
        self.print_calldef_info()

    def visit_casting_operator(self):
        self.print_decl_header()
        self.print_calldef_info()

    def visit_free_function(self):
        self.print_decl_header()
        self.print_calldef_info()

    def visit_free_operator(self):
        self.print_decl_header()
        self.print_calldef_info()

    def visit_class_declaration(self):
        self.print_decl_header()

    def visit_class(self):
        self.print_decl_header()
        curr_level = self.level + 1
        class_type = 'class type: ' + "'%s'" % str(self.__inst.class_type)
        self.writer(
            ' ' *
            curr_level *
            self.INDENT_SIZE +
            class_type.ljust(self.JUSTIFY))
        if self.__print_details:
            byte_size = 'size: %d' % self.__inst.byte_size
            self.writer(
                ' ' *
                curr_level *
                self.INDENT_SIZE +
                byte_size.ljust(self.JUSTIFY))
            try:
                byte_align = 'align: %d' % self.__inst.byte_align
                self.writer(
                    ' ' *
                    curr_level *
                    self.INDENT_SIZE +
                    byte_align.ljust(self.JUSTIFY))
            except NotImplementedError:
                self.writer(
                    ' ' *
                    curr_level *
                    self.INDENT_SIZE +
                    "align: not implemented".ljust(self.JUSTIFY))

        if self.__inst.aliases:
            aliases = sorted([typedef.name for typedef in self.__inst.aliases])
            msg = 'aliases: ' + repr(aliases)
            self.writer(
                ' ' *
                curr_level *
                self.INDENT_SIZE +
                msg.ljust(self.JUSTIFY) +
                os.linesep)

        def print_hierarchy(hierarchy_type, classes, curr_level):
            self.writer(
                ' ' *
                curr_level *
                self.INDENT_SIZE +
                hierarchy_type.ljust(self.JUSTIFY) +
                os.linesep)
            curr_level += 1
            for class_ in classes:
                class_str = 'class: ' + \
                    "'%s'" % str(class_.related_class.decl_string)
                self.writer(
                    ' ' *
                    curr_level *
                    self.INDENT_SIZE +
                    class_str.ljust(self.JUSTIFY) +
                    os.linesep)
                access = 'access type: ' + "'%s'" % str(class_.access)
                self.writer(' ' *
                            (curr_level +
                             1) *
                            self.INDENT_SIZE +
                            access.ljust(self.JUSTIFY) +
                            os.linesep)
                if class_.is_virtual is not None:
                    is_virtual = 'virtual inheritance: ' + \
                        "'%s'" % str(class_.is_virtual)
                    self.writer(' ' *
                                (curr_level +
                                 1) *
                                self.INDENT_SIZE +
                                is_virtual.ljust(self.JUSTIFY) +
                                os.linesep)

        def print_members(members_type, members, curr_level):
            members = sorted(members[:])
            self.writer(
                ' ' *
                curr_level *
                self.INDENT_SIZE +
                members_type.ljust(self.JUSTIFY) +
                os.linesep)
            if self.__recursive:
                curr_level += 1
                for member in members:
                    prn = self.clone()
                    prn.instance = member
                    algorithm.apply_visitor(prn, member)

        if self.__inst.bases:
            print_hierarchy('base classes: ', self.__inst.bases, curr_level)

        if self.__inst.derived:
            print_hierarchy(
                'derived classes: ',
                self.__inst.derived,
                curr_level)

        print_members('public: ', self.__inst.public_members, curr_level)
        print_members('protected: ', self.__inst.protected_members, curr_level)
        print_members('private: ', self.__inst.private_members, curr_level)

    def visit_enumeration(self):
        self.print_decl_header()
        curr_level = self.level + 1
        self.writer(
            ' ' *
            curr_level *
            self.INDENT_SIZE +
            'values:'.ljust(self.JUSTIFY) +
            os.linesep)
        value_level = ' ' * (curr_level + 1) * self.INDENT_SIZE
        self.writer(os.linesep + os.linesep)
        for name, value in self.__inst.values:
            self.writer(value_level + "%s : %s" % (name, value) + os.linesep)

    def visit_namespace(self):
        if not self.verbose and not self.__inst.declarations:
            return  # don't print info about empty namespaces
        self.print_decl_header()
        if self.__recursive:
            inst_decls = sorted(self.__inst.declarations[:])
            for decl in inst_decls:
                if self.is_builtin_decl(decl):
                    continue
                prn = self.clone()
                prn.instance = decl
                algorithm.apply_visitor(prn, decl)

    def visit_typedef(self):
        self.print_decl_header()
        curr_level = self.level + 1
        self.writer(
            ' ' *
            curr_level *
            self.INDENT_SIZE +
            'alias to: ' +
            self.__inst.decl_type.decl_string +
            os.linesep)

    def visit_variable(self):
        self.print_decl_header()
        curr_level = self.level + 1
        self.writer(
            ' ' *
            curr_level *
            self.INDENT_SIZE +
            'type: %s' %
            self.__inst.decl_type.decl_string)
        self.writer(
            ' ' *
            curr_level *
            self.INDENT_SIZE +
            'value: %s' %
            self.__inst.value)
        if self.__print_details:
            if self.__inst.bits:
                bits = 'bits: %d' % self.__inst.bits
                self.writer(
                    ' ' *
                    curr_level *
                    self.INDENT_SIZE +
                    bits.ljust(self.JUSTIFY))

            byte_size = 'size: %d' % self.__inst.decl_type.byte_size
            self.writer(
                ' ' *
                curr_level *
                self.INDENT_SIZE +
                byte_size.ljust(self.JUSTIFY))
            try:
                byte_align = 'align: %d' % self.__inst.decl_type.byte_align
                self.writer(
                    ' ' *
                    curr_level *
                    self.INDENT_SIZE +
                    byte_align.ljust(self.JUSTIFY))
            except NotImplementedError:
                self.writer(
                    ' ' *
                    curr_level *
                    self.INDENT_SIZE +
                    "align: not implemented".ljust(self.JUSTIFY))
            byte_offset = 'offset: %d' % self.__inst.byte_offset
            self.writer(
                ' ' *
                curr_level *
                self.INDENT_SIZE +
                byte_offset +
                os.linesep)


def print_declarations(
        decls,
        detailed=True,
        recursive=True,
        writer=lambda x: sys.stdout.write(x + os.linesep),
        verbose=True):
    """
    print declarations tree rooted at each of the included nodes.

    :param decls: either a single :class:declaration_t object or list
        of :class:declaration_t objects
    """
    prn = decl_printer_t(0, detailed, recursive, writer, verbose=verbose)
    if not isinstance(decls, list):
        decls = [decls]
    for d in decls:
        prn.level = 0
        prn.instance = d
        algorithm.apply_visitor(prn, d)


def dump_declarations(declarations, file_path):
    """
    Dump declarations tree rooted at each of the included nodes to the file

    :param declarations: either a single :class:declaration_t object or a list
        of :class:declaration_t objects
    :param file_path: path to a file

    """

    with open(file_path, "w+") as f:
        print_declarations(declarations, writer=f.write)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import platform
import subprocess

from pygccxml import declarations

from . import linker
from . import config
from . import patcher
from . import declarations_cache
from . import declarations_joiner
from .etree_scanner import ietree_scanner_t as scanner_t

from .. import utils


class source_reader_t(object):
    """
    This class reads C++ source code and returns the declarations tree.

    This class is the only class that works directly with CastXML.

    It has only one responsibility: it calls CastXML with a source file
    specified by the user and creates declarations tree. The implementation of
    this class is split to two classes:

    1. `scanner_t` - this class scans the "XML" file, generated by CastXML
        or CastXML and creates :mod:`pygccxml` declarations and types classes.
        After the XML file has been processed declarations and type class
        instances keeps references to each other using CastXML generated id's.

    2. `linker_t` - this class contains logic for replacing CastXML
        generated ids with references to declarations or type class instances.
    """

    def __init__(self, configuration, cache=None, decl_factory=None):
        """
        :param configuration:
                       Instance of :class:`xml_generator_configuration_t`
                       class, that contains CastXML configuration.

        :param cache: Reference to cache object, that will be updated after a
                      file has been parsed.
        :type cache: Instance of :class:`cache_base_t` class

        :param decl_factory: Declarations factory, if not given default
                             declarations factory( :class:`decl_factory_t` )
                             will be used.

        """

        self.logger = utils.loggers.cxx_parser
        self.__search_directories = []
        self.__config = configuration
        self.__cxx_std = utils.cxx_standard(configuration.cflags)
        self.__search_directories.append(configuration.working_directory)
        self.__search_directories.extend(configuration.include_paths)
        if not cache:
            cache = declarations_cache.dummy_cache_t()
        self.__dcache = cache
        self.__config.raise_on_wrong_settings()
        self.__decl_factory = decl_factory
        if not decl_factory:
            self.__decl_factory = declarations.decl_factory_t()
        self.__xml_generator_from_xml_file = None

    @property
    def xml_generator_from_xml_file(self):
        """
        Configuration object containing information about the xml generator
        read from the xml file.

        Returns:
            utils.xml_generators: configuration object
        """
        return self.__xml_generator_from_xml_file

    def __create_command_line(self, source_file, xml_file):
        """
        Generate the command line used to build xml files.

        """
        return self.__create_command_line_castxml(source_file, xml_file)

    def __create_command_line_common(self):
        assert isinstance(self.__config, config.xml_generator_configuration_t)

        cmd = []

        # Add xml generator executable (between "" for windows)
        cmd.append('"%s"' % os.path.normpath(self.__config.xml_generator_path))

        # Add all passed cflags
        if self.__config.cflags != "":
            cmd.append(" %s " % self.__config.cflags)

        # Add additional includes directories
        dirs = self.__search_directories
        cmd.append(''.join([' -I"%s"' % search_dir for search_dir in dirs]))

        return cmd

    def __create_command_line_castxml(self, source_file, xmlfile):

        cmd = self.__create_command_line_common()

        # Clang option: -c Only run preprocess, compile, and assemble steps
        cmd.append("-c")
        # Clang option: make sure clang knows we want to parse c++
        cmd.append("-x c++")

        # Always require a compiler path at this point
        if self.__config.compiler_path is None:
            raise (RuntimeError(
                "Please pass the compiler_path as argument to " +
                "your xml_generator_configuration_t(), or add it to your " +
                "pygccxml configuration file."))

        # Platform specific options
        if platform.system() == 'Windows':
            compilers = ("mingw", "g++", "gcc")
            compiler_path = self.__config.compiler_path.lower()
            if any(compiler in compiler_path for compiler in compilers):
                # Look at the compiler path. This is a bad way
                # to find out if we are using mingw; but it
                # should probably work in most of the cases
                cmd.append('--castxml-cc-gnu ' + self.__config.compiler_path)
            else:
                # We are using msvc
                cmd.append('--castxml-cc-msvc ' +
                           '"%s"' % self.__config.compiler_path)
                if self.__config.compiler == 'msvc9':
                    cmd.append('"-D_HAS_TR1=0"')
        else:
            # On mac or linux, use gcc or clang (the flag is the same)
            cmd.append('--castxml-cc-gnu ')

            if self.__cxx_std.is_implicit:
                std_flag = ''
            else:
                std_flag = ' ' + self.__cxx_std.stdcxx + ' '

            ccflags = self.__config.ccflags
            if std_flag:
                ccflags += std_flag

            if ccflags:
                all_cc_opts = self.__config.compiler_path + ' ' + ccflags
                cmd.append(
                    '"(" ' + all_cc_opts + ' ")"')
            else:
                cmd.append(self.__config.compiler_path)

        if self.__config.castxml_epic_version is not None:
            if self.__config.castxml_epic_version != 1:
                raise RuntimeError(
                    "The CastXML epic version can only be 1, "
                    "but it was " + str(self.__config.castxml_epic_version))
            # Tell castxml to output xml file with a specific epic version
            cmd.append(
                '--castxml-output=' + str(self.__config.castxml_epic_version))
        else:
            # Tell castxml to output xml files that are backward compatible
            # with the format from gccxml
            cmd.append('--castxml-gccxml')

        # Add symbols
        cmd = self.__add_symbols(cmd)

        # The destination file
        cmd.append('-o %s' % xmlfile)
        # The source file
        cmd.append('%s' % source_file)
        # Where to start the parsing
        if self.__config.start_with_declarations:
            cmd.append(
                '--castxml-start "%s"' %
                ','.join(self.__config.start_with_declarations))
        cmd_line = ' '.join(cmd)
        self.logger.debug('castxml cmd: %s', cmd_line)
        return cmd_line

    def __add_symbols(self, cmd):
        """
        Add all additional defined and undefined symbols.

        """

        if self.__config.define_symbols:
            symbols = self.__config.define_symbols
            cmd.append(''.join(
                [' -D"%s"' % def_symbol for def_symbol in symbols]))

        if self.__config.undefine_symbols:
            un_symbols = self.__config.undefine_symbols
            cmd.append(''.join(
                [' -U"%s"' % undef_symbol for undef_symbol in un_symbols]))

        return cmd

    def create_xml_file(self, source_file, destination=None):
        """
        This method will generate a xml file using an external tool.

        The method will return the file path of the generated xml file.

        :param source_file: path to the source file that should be parsed.
        :type source_file: str

        :param destination: if given, will be used as target file path for
                            the xml generator.
        :type destination: str

        :rtype: path to xml file.

        """

        xml_file = destination
        # If file specified, remove it to start else create new file name
        if xml_file:
            utils.remove_file_no_raise(xml_file, self.__config)
        else:
            xml_file = utils.create_temp_file_name(suffix='.xml')

        ffname = source_file
        if not os.path.isabs(ffname):
            ffname = self.__file_full_name(source_file)
        command_line = self.__create_command_line(ffname, xml_file)

        process = subprocess.Popen(
            args=command_line,
            shell=True,
            stdout=subprocess.PIPE)

        try:
            results = []
            while process.poll() is None:
                line = process.stdout.readline()
                if line.strip():
                    results.append(line.rstrip())
            for line in process.stdout.readlines():
                if line.strip():
                    results.append(line.rstrip())

            exit_status = process.returncode
            msg = os.linesep.join([str(s) for s in results])
            if self.__config.ignore_gccxml_output:
                if not os.path.isfile(xml_file):
                    raise RuntimeError(
                        "Error occurred while running " +
                        self.__config.xml_generator.upper() +
                        ": %s status:%s" %
                        (msg, exit_status))
            else:
                if msg or exit_status or not \
                        os.path.isfile(xml_file):
                    if not os.path.isfile(xml_file):
                        raise RuntimeError(
                            "Error occurred while running " +
                            self.__config.xml_generator.upper() +
                            " xml file does not exist")
                    else:
                        raise RuntimeError(
                            "Error occurred while running " +
                            self.__config.xml_generator.upper() +
                            ": %s status:%s" % (msg, exit_status))
        except Exception:
            utils.remove_file_no_raise(xml_file, self.__config)
            raise
        finally:
            process.wait()
            process.stdout.close()
        return xml_file

    def create_xml_file_from_string(self, content, destination=None):
        """
        Creates XML file from text.

        :param content: C++ source code
        :type content: str

        :param destination: file name for xml file
        :type destination: str

        :rtype: returns file name of xml file
        """
        header_file = utils.create_temp_file_name(suffix='.h')

        try:
            with open(header_file, "w+") as header:
                header.write(content)
            xml_file = self.create_xml_file(header_file, destination)
        finally:
            utils.remove_file_no_raise(header_file, self.__config)
        return xml_file

    def read_file(self, source_file):
        return self.read_cpp_source_file(source_file)

    def read_cpp_source_file(self, source_file):
        """
        Reads C++ source file and returns declarations tree

        :param source_file: path to C++ source file
        :type source_file: str

        """

        xml_file = ''
        try:
            ffname = self.__file_full_name(source_file)
            self.logger.debug("Reading source file: [%s].", ffname)
            decls = self.__dcache.cached_value(ffname, self.__config)
            if not decls:
                self.logger.debug(
                    "File has not been found in cache, parsing...")
                xml_file = self.create_xml_file(ffname)
                decls, files = self.__parse_xml_file(xml_file)
                self.__dcache.update(
                    ffname, self.__config, decls, files)
            else:
                self.logger.debug((
                    "File has not been changed, reading declarations " +
                    "from cache."))
        except Exception:
            if xml_file:
                utils.remove_file_no_raise(xml_file, self.__config)
            raise
        if xml_file:
            utils.remove_file_no_raise(xml_file, self.__config)

        return decls

    def read_xml_file(self, xml_file):
        """
        Read generated XML file.

        :param xml_file: path to xml file
        :type xml_file: str

        :rtype: declarations tree

        """

        assert self.__config is not None

        ffname = self.__file_full_name(xml_file)
        self.logger.debug("Reading xml file: [%s]", xml_file)
        decls = self.__dcache.cached_value(ffname, self.__config)
        if not decls:
            self.logger.debug("File has not been found in cache, parsing...")
            decls, _ = self.__parse_xml_file(ffname)
            self.__dcache.update(ffname, self.__config, decls, [])
        else:
            self.logger.debug(
                "File has not been changed, reading declarations from cache.")

        return decls

    def read_string(self, content):
        """
        Reads a Python string that contains C++ code, and return
        the declarations tree.

        """

        header_file = utils.create_temp_file_name(suffix='.h')
        with open(header_file, "w+") as f:
            f.write(content)

        try:
            decls = self.read_file(header_file)
        except Exception:
            utils.remove_file_no_raise(header_file, self.__config)
            raise
        utils.remove_file_no_raise(header_file, self.__config)

        return decls

    def __file_full_name(self, file_):
        if os.path.isfile(file_):
            return file_
        for path in self.__search_directories:
            file_path = os.path.join(path, file_)
            if os.path.isfile(file_path):
                return file_path
        raise RuntimeError("pygccxml error: file '%s' does not exist" % file_)

    def __parse_xml_file(self, xml_file):
        scanner_ = scanner_t(xml_file, self.__decl_factory, self.__config)
        scanner_.read()
        self.__xml_generator_from_xml_file = \
            scanner_.xml_generator_from_xml_file
        decls = scanner_.declarations()
        types = scanner_.types()
        files = scanner_.files()
        linker_ = linker.linker_t(
            decls=decls,
            types=types,
            access=scanner_.access(),
            membership=scanner_.members(),
            files=files,
            xml_generator_from_xml_file=self.__xml_generator_from_xml_file)
        for type_ in list(types.values()):
            # I need this copy because internaly linker change types collection
            linker_.instance = type_
            declarations.apply_visitor(linker_, type_)
        for decl in decls.values():
            linker_.instance = decl
            declarations.apply_visitor(linker_, decl)
        declarations_joiner.bind_aliases(iter(decls.values()))

        # Patch the declarations tree
        if self.__xml_generator_from_xml_file.is_castxml:
            patcher.update_unnamed_class(decls.values())
        patcher.fix_calldef_decls(
            scanner_.calldefs(), scanner_.enums(), self.__cxx_std)

        decls = [inst for inst in iter(decls.values()) if self.__check(inst)]
        return decls, list(files.values())

    @staticmethod
    def __check(inst):
        return isinstance(inst, declarations.namespace_t) and not inst.parent



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines class :class:`mdecl_wrapper_t` that allows to work on set of
declarations, as it was one declaration.

The :class:`class <mdecl_wrapper_t>` allows user to not write "for" loops
within the code.
"""

import os


class call_redirector_t(object):

    """Internal class used to call some function of objects"""

    def __init__(self, name, decls):
        """creates call_redirector_t instance.

        :param name: name of method, to be called on every object in the
        `decls` list
        :param decls: list of objects
        """
        object.__init__(self)
        self.name = name
        self.decls = decls

    def __call__(self, *arguments, **keywords):
        """calls method :attr:`call_redirector_t.name` on every object
        within the :attr:`call_redirector_t.decls` list"""
        for d in self.decls:
            callable_ = getattr(d, self.name)
            callable_(*arguments, **keywords)


class mdecl_wrapper_t(object):

    """
    multiple declarations class wrapper

    The main purpose of this class is to allow an user to work on many
    declarations, as they were only one single declaration.

    For example, instead of writing `for` loop like the following

    .. code-block:: python

       for c in global_namespace.classes():
           c.attribute = "xxxx"

    you can write:

    .. code-block:: python

       global_namespace.classes().attribute = "xxxx"

    The same functionality could be applied on "set" methods too.
    """

    def __init__(self, decls):
        """:param decls: list of declarations to operate on.
        :type decls: list of :class:`declaration wrappers <decl_wrapper_t>`
        """
        object.__init__(self)
        self.__dict__['declarations'] = decls

    def __bool__(self):
        return bool(self.declarations)

    def __len__(self):
        """returns the number of declarations"""
        return len(self.declarations)

    def __getitem__(self, index):
        """provides access to declaration"""
        return self.declarations[index]

    def __iter__(self):
        return iter(self.declarations)

    def __ensure_attribute(self, name):
        invalid_decls = [d for d in self.declarations if not hasattr(d, name)]
        sep = os.linesep + '    '
        if invalid_decls:
            raise RuntimeError(
                "Next declarations don't have '%s' attribute: %s" %
                (name, sep.join(map(str, invalid_decls))))

    def __setattr__(self, name, value):
        """Updates the value of attribute on all declarations.
        :param name: name of attribute
        :param value: new value of attribute
        """
        self.__ensure_attribute(name)
        for d in self.declarations:
            setattr(d, name, value)

    def __getattr__(self, name):
        """:param name: name of method
        """
        return call_redirector_t(name, self.declarations)

    def __contains__(self, item):
        return item in self.declarations
